import requests
import json
import os
import re
try: # Python 2/3 compatibility
    import urllib.request as urlrequest
except ImportError:
    import urllib as urlrequest

def help():
    print('- makeAuthHeader(apiKey)')
    print('Returns an HTTP Authorization header entry (in dictionary format) created from the provided API key. The API key can be accessed from the settings page in Studio.')
    print('')
    print('- getResults(taskId, baseUrl, authHeader, incIncompleteInstances=\'false\', incIterations=\'true\')')
    print('Make an API call to get a list of results from completed instances or iterations. Task ID is available from Studio after clicking on the relevant task. The baseUrl string is usually: \nr\'https://studio.lambda.hvmd.io/\'\n The authHeader is a dictionary specifying an HTTP Authorisation header; it can be created with makeAuthHeader. Each entry in the list is a dictionary containing the results of one instance, in the structure shown below.')
    print('')
    print('- getInstances(taskId, baseUrl, authHeader)')
    print('Make an API call to get a list of instances submitted to the task. Task ID is available from Studio after clicking on the relevant task. The baseUrl string is usually: \nr\'https://studio.lambda.hvmd.io/\'\n The authHeader is a dictionary specifying an HTTP Authorisation header; it can be created with makeAuthHeader. Each entry in the list is a dictionary containing the input data for one instance, in the structure shown below.')
    print('')
    print('- transferEntry(source, target, entryField="data", sourceIdField="id", targetIdField="instanceId")')
    print('Transfer an entry from one list of dictionaries to another list of dictionaries. Used for copying input data from instance dictionaries to results dictionaries, matching on instanceId.')
    print('')
    print('- getInstancesAndResults(taskId, baseUrl, apiKey, incIncompleteInstances=\'false\', incIterations=\'true\')')
    print('Call both getResults and getInstances, and stitch them together with transferEntry')
    print('')
    print('- listFilesUploaded(taskId, baseUrl, apiKey)')
    print('Make an API call to get a list of all uploaded files')
    print('')
    print('- uploadFiles(taskId, baseUrl, apiKey, lstFilenames, localRepo, overwrite=False)')
    print('Upload local files from localRepo with filenames in lstFilenames to the specified task')
    print('')
    print('- downloadFile(filepath, localRepo, prefix=\'\', overwrite=False)')
    print('Download a file generated by a task from the long s3 link provided as output. Provide a prefix (iteration ID etc). Returns error code (0=success) and file name.')
    print('')
    print('- submitInstance(taskId, baseUrl, apiKey, name, instructions='', data={}, overrideSchema={}, tags=[])')
    print('Make an API call to create an instance. \'Name\' is a compulsory string. \'Instructions\' is the markdown shown to the contributor for the instance. \'Data\' is a dictionary.')
    print('')
    print('')
    print('- Instance Structure')
    print('Each entry in the list returned by getInstances() is a dictionary with the following entries:')
    print('id [instance id]\nname\ndata [a dictionary]\nstatus\ntags [a list of strings]\nsummary\ncreated')
    print('')
    print('- Results Structure')
    print('Each entry in the list returned by getResults() is a dictionary with the following entries:')
    print('instanceId\nstatus\nlastUpdated\nagreement\ntags [a list of strings]\nresult [a dictionary with entries \'data\' and \'isError\']\niterationResults [a list of entries in the structure below]')
    print('Note that \'result\' is the output of error checking and will be empty if that is off, even with only one iteration.')
    print('')
    print('- Iteration Structure')
    print('Each entry in the list of iterationResults above is a dictionary with the following entries:')
    print('worker\nstart\nend\niterationId\ndata [a dictionary of the collected data keyed by input box names]\nisError')
 

def makeAuthHeader(apiKey):
    return {'Authorization': 'ApiKey ' + apiKey}

def getResults(taskId, baseUrl, apiKey, incIncompleteInstances='false', incIterations='true'):  
    # Probably unnecessary but just in case
    if(incIncompleteInstances==True):
        incIncompleteInstances='true'
    if(incIncompleteInstances==False):
        incIncompleteInstances='false'
    if(incIterations==True):
        incIterations='true'
    if(incIterations==False):
        incIterations='false'
        
    authHeader = makeAuthHeader(apiKey)
    with requests.Session() as s:
        s.headers.update(authHeader)
        
        paginatedUrl = '{0}/api/tasks/{1}/results?perPage=1000&incIncompleteInstances={2}&incIterations={3}'.format(baseUrl, taskId, incIncompleteInstances, incIterations)
        results = []
        print(paginatedUrl)

        while paginatedUrl is not None:            
            response = s.get(paginatedUrl)              
            results += response.json()
            n = response.links["next"] if "next" in response.links else None
            if n is not None:
                paginatedUrl = baseUrl + n["url"]
            else:
                paginatedUrl = None
    return results    

def getInstances(taskId, baseUrl, apiKey): 
    authHeader = makeAuthHeader(apiKey)
    with requests.Session() as s:
        s.headers.update(authHeader)
        
        paginatedUrl = '{0}/api/tasks/{1}/instances?perPage=1000'.format(baseUrl, taskId)
        instances = []

        while paginatedUrl is not None:
            response = s.get(paginatedUrl) 
            instances += response.json()
            n = response.links["next"] if "next" in response.links else None
            if n is not None:
                paginatedUrl = baseUrl + n["url"]
            else:
                paginatedUrl = None
    return instances       

def getInstanceIds(taskId, baseUrl, apiKey):
    instances = getInstances(taskId, baseUrl, apiKey)
    instanceIds = []
    for i in instances:
        print(i)
        instanceIds.append(i.get('id'))
    return instanceIds

def transferEntry(source, target, entryField="data", sourceIdField="id", targetIdField="instanceId"):
    # source and target are each an array of dictionaries.
    # For each entry in source, extract "entry" and insert it into  
    # the dictionary in target in which sourceId matches targetId.
    # By default transfers "data" from list of instance dictionaries to list of results dictionaries.
    # If entry doesn't exist in source, insert null into target
    
    # Create set of sourceId : entry pairs
    sourceIdEntryPairs = {}
    for s in source:
        entry = s.get(entryField)
        sourceId = s.get(sourceIdField)
        sourceIdEntryPairs[sourceId] = entry
    for t in target:
        targetId = t.get(targetIdField)
        t[entryField] = sourceIdEntryPairs.get(targetId)

def getInstancesAndResults(taskId, baseUrl, apiKey, incIncompleteInstances='false', incIterations='true'):
    #Make API calls for both instances and results, and transfer data from the former to the latter.
    instances = getInstances(taskId, baseUrl, apiKey)
    results = getResults(taskId, baseUrl, apiKey, incIncompleteInstances, incIterations)
    transferEntry(instances, results)
    return results

def listFilesUploaded(taskId, baseUrl, apiKey):
    # Make API call to get list of filenames uploaded to the task.
    authHeader = makeAuthHeader(apiKey)
    with requests.Session() as s:
        s.headers.update(authHeader)
        paginatedUrl = '{0}/api/tasks/{1}/files?perPage=1000'.format(baseUrl, taskId)
        result = []

        while paginatedUrl is not None:
            response = s.get(paginatedUrl)
            result += response.json()
            n = response.links["next"] if "next" in response.links else None
            if n is not None:
                paginatedUrl = HivemindUrl + n["url"]
            else:
                paginatedUrl = None
        lstFilesAlreadyUploaded = map(lambda x: x.get('filename'), result)
    return list(lstFilesAlreadyUploaded)

def uploadFiles(taskId, baseUrl, apiKey, lstFilenames, localRepo, overwrite=False):    
    if(overwrite==False):
        filesUploaded = listFilesUploaded(taskId, baseUrl,apiKey)
        lstFilenames = list(set(lstFilenames) - set(filesUploaded))
        
    authHeader = {'Authorization' : 'ApiKey {}'.format(apiKey)}
    with requests.Session() as s:
        s.headers.update(authHeader)
        fileUrl = '{0}/api/tasks/{1}/files'.format(baseUrl, taskId)
        files = map(lambda x: os.path.join(localRepo, x), lstFilenames)

        for filename in files:
            with open(filename, 'rb') as f:
                try:
                    response = s.post(fileUrl, files = {'file' : f})
                except:
                    continue

def downloadFile(filepath, localRepo, prefix='', overwrite=False):

    pattern = r'/hivemind\.lambda\.task\.\d+/output/\d+/\d+/(.*)\?X-Amz-Expires='
    filename = re.search(pattern, filepath).group(1)

    newFilename = '{0}_{1}'.format(prefix, filename) if prefix!='' else filename

    lstFilesDownloaded = os.listdir(localRepo)
    if overwrite or (newFilename not in lstFilesDownloaded):
        urlrequest.urlretrieve(filepath, os.path.join(localRepo, newFilename))

    lstFilesDownloaded = os.listdir(localRepo)
    if newFilename in lstFilesDownloaded:
        return [0, newFilename]
    else:
         return [1, newFilename]


def deleteIteration(taskId, baseUrl, apiKey, instanceId, iterationId):    
    authHeader = {'Authorization' : 'ApiKey {}'.format(apiKey)}    
    with requests.Session() as s:
        s.headers.update(authHeader)
        response = s.delete('{0}/api/tasks/{1}/instances/{2}/iterations/{3}'.format(HivemindUrl, taskId, instanceId, iterationId))        

def createInstance(taskId, baseUrl, apiKey, name, instructions='', data={}, overrideSchema={}, tags=[]):
    authHeader = {'Authorization' : 'ApiKey {}'.format(apiKey)} 
    with requests.Session() as s:
        s.headers.update(authHeader)
               
        instance = {
            'name' : name,
            'instruction' : instructions,
            'data' : data,
            'overrideSchema' : overrideSchema,
            'tags' : tags
        }        
        createInstance = s.post('{0}/api/tasks/{1}/instances'.format(baseUrl, taskId), json = instance)


